"""
@Author: WangYuXiang
@E-mile: Hill@3io.cc
@CreateTime: 2021/1/19 15:44
@DependencyLibrary:
@MainFunction：
@FileDoc:
    login.py
    基础视图文件
    BaseView    只实现路由分发的基础视图
    GeneralView 通用视图，可以基于其实现增删改查，提供权限套件
    ViewSetView 视图集视图，可以配合Mixin实现复杂的视图集，
                数据来源基于模型查询集,可以配合Route组件实现便捷的路由管理



"""
from datetime import datetime

from sanic.log import logger
from sanic.response import json, HTTPResponse
from simplejson import dumps
from tortoise.exceptions import IntegrityError
from tortoise.transactions import in_transaction

from srf.constant import ALL_METHOD, DEFAULT_METHOD_MAP
from srf.exceptions import APIException, ValidationException
from srf.permissions import ViewMapPermission
from srf.status import RuleStatus, HttpStatus
from srf import mixins

__all__ = ['BaseView', 'GeneralView', 'ViewSetView', 'ModelViewSet']

from srf.utils import run_awaitable


class BaseView:
    """只实现路由分发的基础视图
    在使用时应当开放全部路由 ALL_METHOD
    app.add_route('/test', BaseView.as_view(), 'test', ALL_METHOD)
    如需限制路由则在其他地方注明
    app.add_route('/test', BaseView.as_view(), 'test', ALL_METHOD)
    注意以上方法的报错是不可控的
    """

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    @classmethod
    def as_view(cls, *class_args, **class_kwargs):
        # 返回的响应方法闭包
        def view(request, *args, **kwargs):
            self = view.base_class(*class_args, **class_kwargs)

            self.request = request
            self.args = args
            self.kwargs = kwargs
            self.app = request.app
            return self.dispatch(request, *args, **kwargs)

        view.base_class = cls
        view.API_DOC_CONFIG = class_kwargs.get('API_DOC_CONFIG')  # 未来的API文档配置属性+
        view.__doc__ = cls.__doc__
        view.__module__ = cls.__module__
        view.__name__ = cls.__name__
        return view
    
    async def dispatch(self, request, *args, **kwargs):
        """分发路由"""
        request.user = None
        method = request.method.lower()
        
        if not hasattr(self,method):
            return HTTPResponse('405请求方法错误', status=405)
        handler = getattr(self, method, None)
        return await run_awaitable(handler,request, *args, **kwargs)


class APIView(BaseView):
    """通用视图，可以基于其实现增删改查，提供权限套件"""
    authentication_classes = ()
    permission_classes = (ViewMapPermission,)
    is_transaction = True

    async def dispatch(self, request, *args, **kwargs):
        """分发路由"""
        request.user = None
        method = request.method.lower()
        
        if not hasattr(self,method):
            return self.json_response(msg='发生错误：未找到%s方法' % method, status=RuleStatus.STATUS_0_FAIL,
                                      http_status=HttpStatus.HTTP_405_METHOD_NOT_ALLOWED)
        handler = getattr(self, method, None)
        try:
            await self.initial(request, *args, **kwargs)
            if self.is_transaction:
                async with in_transaction():
                    response = await run_awaitable(handler, request=request, *args, **kwargs)
            else:
                response = await run_awaitable(handler, request=request, *args, **kwargs)
        except APIException as exc:
            response = self.handle_exception(exc)
        except ValidationException as exc:
            response = self.error_json_response(exc.error_detail, '数据验证失败')
        except AssertionError as exc:
            raise exc
        except IntegrityError as exc:
            response = self.error_json_response(msg=str(exc), status=RuleStatus.STATUS_0_FAIL)
        except Exception as exc:
            logger.error('--捕获未知错误--', exc)
            msg = '发生致命的未知错误，请在服务器查看时间为{}的日志'.format(datetime.now().strftime('%F %T'))
            response = self.json_response(msg=msg, status=RuleStatus.STATUS_0_FAIL,
                                          http_status=HttpStatus.HTTP_500_INTERNAL_SERVER_ERROR)
        return response

    def handle_exception(self, exc: APIException):
        return self.json_response(**exc.response_data())

    def json_response(self, data=None, msg="OK", status=RuleStatus.STATUS_1_SUCCESS,
                      http_status=HttpStatus.HTTP_200_OK):
        """
        Json 相应体
        :param data: 返回的数据主题
        :param msg: 前台提示字符串
        :param status: 前台约定状态，供前台判断是否成功
        :param http_status: Http响应数据
        :return:
        """
        if data is None:
            data = {}
        response_body = {
            'data': data,
            'message': msg,
            'status': status
        }
        return json(body=response_body, status=http_status, dumps=dumps)

    def success_json_response(self, data=None, msg="Success", **kwargs):
        """
        快捷的成功的json响应体
        :param data: 返回的数据主题
        :param msg: 前台提示字符串
        :return: json
        """
        status = kwargs.pop('status', RuleStatus.STATUS_1_SUCCESS)
        http_status = kwargs.pop('http_status', HttpStatus.HTTP_200_OK)
        return self.json_response(data=data, msg=msg, status=status, http_status=http_status)

    def error_json_response(self, data=None, msg="Fail", **kwargs):
        """
        快捷的失败的json响应体
        :param data: 返回的数据主题
        :param msg: 前台提示字符串
        :return: json
        """
        status = kwargs.pop('status', RuleStatus.STATUS_0_FAIL)
        http_status = kwargs.pop('http_status', HttpStatus.HTTP_400_BAD_REQUEST)
        return self.json_response(data=data, msg=msg, status=status, http_status=http_status)

    def get_authenticators(self):
        """
        实例化并返回此视图可以使用的身份验证器列表
        """
        return [auth() for auth in self.authentication_classes]

    async def check_authentication(self, request):
        """
        检查权限 查看是否拥有权限，并在此处为Request.User 赋值
        :param request: 请求
        :return:
        """
        for authenticators in self.get_authenticators():
            await authenticators.authenticate(request)

    def get_permissions(self):
        """
        实例化并返回此视图所需的权限列表
        """
        return [permission() for permission in self.permission_classes]

    async def check_permissions(self, request):
        """
        检查是否应允许该请求，如果不允许该请求，
        则在 has_permission 中引发一个适当的异常。
        :param request: 当前请求
        :return:
        """
        for permission in self.get_permissions():
            await permission.has_permission(request, self)

    async def check_object_permissions(self, request, obj):
        """
        检查是否应允许给定对象的请求, 如果不允许该请求，
        则在 has_object_permission 中引发一个适当的异常。
            常用于 get_object() 方法
        :param request: 当前请求
        :param obj: 需要鉴权的模型对象
        :return:
        """
        for permission in self.get_permissions():
            await permission.has_object_permission(request, self, obj)

    async def check_throttles(self, request):
        """
        检查范围频率。
        则引发一个 APIException 异常。
        :param request:
        :return:
        """
        pass

    async def initial(self, request, *args, **kwargs):
        """
        在请求分发之前执行初始化操作，用于检查权限及检查基础内容
        """
        await self.check_authentication(request)
        await self.check_permissions(request)
        await self.check_throttles(request)





"""
@Author：WangYuXiang
@E-mile：Hill@3io.cc
@CreateTime：2021/4/25 16:51
@DependencyLibrary：无
@MainFunction：无
@FileDoc： 
    permissions.py
    文件说明
@ChangeHistory:
    datetime action why
    example:
    2021/4/25 16:51 change 'Fix bug'
        
"""
from srf.exceptions import PermissionDenied
from srf.constant import ALL_METHOD


class BasePermission:
    async def has_permission(self, request, view):
        pass

    async def has_obj_permission(self, request, view, obj):
        pass


class ViewMapPermission(BasePermission):
    """
    仅供GeneralView及其子类视图的权限套件

    :param BasePermission: BasePermission
    """
    async def has_permission(self, request, view):
        permission_map = self.get_permission_map(view)
        method = request.method.lower()
        permissions = permission_map.get(method)
        if not request.user.has_permissions(permissions):
            raise PermissionDenied()

    def get_permission_map(self, view):
        permission_map = self.permission_map
        if hasattr(view, 'permission_map'):
            permission_map.update(view.permission_map)
        return permission_map

    @property
    def permission_map(self):
        permission_map = {method.lower(): () for method in ALL_METHOD}
        permission_map['all'] = ()
        return permission_map



"""
@Author：WangYuXiang
@E-mile：Hill@3io.cc
@CreateTime：2021/3/31 16:21
@DependencyLibrary：无
@MainFunction：无
@FileDoc： 
    authentication.py
    文件说明
@ChangeHistory:
    datetime action why
    example:
    2021/3/31 16:21 change 'Fix bug'
        
"""
import jwt
from jwt.exceptions import ExpiredSignatureError

from srf.exceptions import APIException
from srf.request import SRFRequest
from srf.status import HttpStatus


class BaseAuthenticate:
    def authenticate(self, request: SRFRequest, **kwargs):
        """验证权限并返回User对象"""
        # request.headers['']


class BaseTokenAuthenticate(BaseAuthenticate):
    """基于Token的基础验证 JWT """
    token_key = 'X-Token'

    async def authenticate(self, request: SRFRequest, **kwargs):
        """验证逻辑"""
        token = request.headers.get(self.token_key)
        if token is None:
            raise APIException(message='授权错误:请求头{}不存在'.format(self.token_key), http_status=HttpStatus.HTTP_401_UNAUTHORIZED)
        token_secret = request.app.config.TOKEN_SECRET
        try:
            token_info = self.authentication_token(token, token_secret)
        except ExpiredSignatureError:
            raise APIException(message='授权已过期,请重新登录', http_status=HttpStatus.HTTP_401_UNAUTHORIZED)
        await self._authenticate(request, token_info, **kwargs)

    async def _authenticate(self, request: SRFRequest, token_info: dict, **kwargs):
        """主要处理逻辑"""
        pass

    def authentication_token(self, token, token_secret):
        """
        解包Token
        :param token: 口令
        :param token_secret: 解密秘钥
        :return:
        """
        token_info = jwt.decode(token, token_secret, algorithms=['HS256'])
        return token_info



"""
@Author：WangYuXiang
@E-mile：Hill@3io.cc
@CreateTime：2021/1/28 16:16
@DependencyLibrary：无
@MainFunction：无
@FileDoc： 
    test_char_field.py
    字符字段单元测试
@ChangeHistory:
    datetime action why
    example:
    2021/1/28 16:16 change 'Fix bug'
        
"""
import asyncio
import unittest

from tortoise.contrib.test import initializer

from sanic_rest_framework.exceptions import ValidationError
from sanic_rest_framework.fields import CharField
from sanic_rest_framework.test.test_fields.test_base_field import TestBaseField

initializer(['sanic_rest_framework.test.models', ],
            # db_url="sqlite://./db.sqlite",
            loop=asyncio.get_event_loop())


class TestCharField(TestBaseField):
    def test_external_to_internal(self):
        data = ' Python'
        char1 = CharField()
        self.assertEqual(char1.external_to_internal(data), 'Python')

    async def test_internal_to_external(self):
        data1 = {'char1': 'Python'}
        data2 = {'char1': 66666}
        char1 = CharField()
        char1.bind('char1', char1)

        value = await char1.get_internal_value(data1)
        self.assertEqual(await char1.internal_to_external(value), 'Python')

        value = await char1.get_internal_value(data2)
        self.assertEqual(await char1.internal_to_external(value), '66666')

    def test_trim_whitespace(self):
        data = ' Python'
        char1 = CharField()
        char2 = CharField(trim_whitespace=True)
        char3 = CharField(trim_whitespace=False)
        c1_data = char1.external_to_internal(data)
        c2_data = char2.external_to_internal(data)
        c3_data = char3.external_to_internal(data)
        self.assertEqual(c1_data, 'Python')
        self.assertEqual(c2_data, 'Python')
        self.assertEqual(c3_data, ' Python')

    def test_max_length(self):
        data = 'Python'
        char1 = CharField()
        char2 = CharField(max_length=10)
        char3 = CharField(max_length=5)
        self.assertEqual(char1.run_validation(data), 'Python')
        self.assertEqual(char2.run_validation(data), 'Python')

        with self.assertRaises(ValidationError):
            char3.run_validation(data)

    def test_min_length(self):
        data = 'Python'
        char1 = CharField()
        char2 = CharField(min_length=5)
        char3 = CharField(min_length=10)
        self.assertEqual(char1.run_validation(data), 'Python')
        self.assertEqual(char2.run_validation(data), 'Python')

        with self.assertRaises(ValidationError):
            char3.run_validation(data)


if __name__ == '__main__':
    unittest.main()



"""
@Author：WangYuXiang
@E-mile：Hill@3io.cc
@CreateTime：2021/1/29 10:23
@DependencyLibrary：无
@MainFunction：无
@FileDoc：
    test_decimal_field.py
    测试十进制类型字段
@ChangeHistory:
    datetime action why
    example:
    2021/1/29 10:23 change 'Fix bug'

"""

import unittest
from decimal import Decimal, ROUND_DOWN, getcontext, InvalidOperation
from sanic_rest_framework.exceptions import ValidationError
from sanic_rest_framework.fields import DecimalField as TestField
from sanic_rest_framework.test.test_fields.test_base_field import TestBaseField


class TestDecimalField(TestBaseField):

    def test_validate_precision(self):
        """
        测试 max_digits, decimal_places
        :return:
        """

        tf = TestField(max_digits=6, decimal_places=2)
        with self.assertRaises(ValidationError):
            tf.validate_precision(Decimal('99.999'))
        with self.assertRaises(ValidationError):
            tf.validate_precision(Decimal('99999.99'))
        self.assertEqual(tf.validate_precision(Decimal('99.99')), Decimal('99.99'))
        self.assertEqual(tf.validate_precision(Decimal('9999.99')), Decimal('9999.99'))
        with self.assertRaises(ValidationError):
            tf.validate_precision(Decimal('999.999'))

    def test_quantize(self):
        tf = TestField(max_digits=6, decimal_places=2)
        context = getcontext()  # 获取decimal现在的上下文
        context.rounding = ROUND_DOWN
        self.assertEqual(tf.quantize(Decimal('99.99999999')), Decimal('99.99'))

    def test_external_to_internal(self):
        """
        外转内 str -> dict 是严格的，不符合类型的都应该报错,
        一切都要经过验证
        :return:
        """
        tf1 = TestField(max_digits=6, decimal_places=2)

        # 布尔类型 不行
        with self.assertRaises(ValidationError):
            tf1.external_to_internal(self.bool_True)
        # 字符类型 中文 不行
        with self.assertRaises(ValidationError):
            tf1.external_to_internal(self.str_chinese)
        # 字符类型 小数π 超长不行
        with self.assertRaises(ValidationError):
            tf1.external_to_internal(self.str_pi)
        # 浮点类型 小数π 超长不行
        with self.assertRaises(ValidationError):
            self.assertEqual(tf1.external_to_internal(self.pi), self.pi)

        self.assertEqual(tf1.external_to_internal(self.str_max_float), Decimal(self.str_max_float))

        self.assertEqual(tf1.external_to_internal(self.max_int), Decimal(self.max_int))

        self.assertEqual(tf1.external_to_internal(self.str_max_int), Decimal(self.max_int))

    async def test_internal_to_external(self):
        """
        内转外 str -> dict 是宽松的，
        只要类型正确都不报错
        :return:
        """
        data1 = {'tf': self.str_chinese}
        data2 = {'tf': self.max_int}
        data3 = {'tf': self.str_max_int}
        data4 = {'tf': self.str_max_float}
        data5 = {'tf': self.max_float}

        tf = TestField(max_digits=6, decimal_places=2, coerce_to_string=False)
        tf.bind('tf', tf)

        value = await tf.get_internal_value(data1)
        with self.assertRaises(InvalidOperation):
            await tf.internal_to_external(value)

        value = await tf.get_internal_value(data2)
        self.assertEqual(await tf.internal_to_external(value), 9999.00)

        # Decimal(int) == int
        # Decimal(str_int) == int
        value = await tf.get_internal_value(data3)
        self.assertEqual(await tf.internal_to_external(value), self.max_int)

        # Decimal(float) == float
        # Decimal(str_float) != float
        value = await tf.get_internal_value(data4)
        self.assertNotEqual(await tf.internal_to_external(value), self.max_float)
        self.assertEqual(await tf.internal_to_external(value), Decimal(self.str_max_float))

        value = await tf.get_internal_value(data5)
        self.assertEqual(await tf.internal_to_external(value), Decimal(self.str_max_float))

    def test_max_value(self):
        tf1 = TestField(max_digits=6, decimal_places=2)
        tf2 = TestField(max_digits=6, decimal_places=2, max_value=10.69)

        # 未设置 不存在超出限制
        self.assertEqual(tf1.run_validators(data=self.str_max_int), None)
        self.assertEqual(tf1.run_validators(data=self.min_float), None)
        self.assertEqual(tf1.run_validators(data=self.str_max_float), None)
        self.assertEqual(tf1.run_validators(data=self.max_float), None)

        # 超出限制
        with self.assertRaises(ValidationError):
            self.assertEqual(tf2.run_validators(data=self.max_int), None)
        with self.assertRaises(ValidationError):
            self.assertEqual(tf2.run_validators(data=self.max_float), None)
        with self.assertRaises(ValidationError):
            self.assertEqual(tf2.run_validators(data=self.str_max_int), None)
        with self.assertRaises(ValidationError):
            self.assertEqual(tf2.run_validators(data=self.str_max_float), None)

        self.assertEqual(tf2.run_validators(data=self.min_int), None)
        self.assertEqual(tf2.run_validators(data=self.min_float), None)

    def test_min_value(self):
        tf1 = TestField(max_digits=6, decimal_places=2)
        tf2 = TestField(max_digits=6, decimal_places=2, min_value=10.69)

        # 未设置 不存在超出限制
        self.assertEqual(tf1.run_validators(data=self.str_max_int), None)
        self.assertEqual(tf1.run_validators(data=self.min_float), None)
        self.assertEqual(tf1.run_validators(data=self.str_max_float), None)
        self.assertEqual(tf1.run_validators(data=self.max_float), None)

        # 超出限制
        with self.assertRaises(ValidationError):
            self.assertEqual(tf2.run_validators(data=self.min_float), None)
        with self.assertRaises(ValidationError):
            self.assertEqual(tf2.run_validators(data=self.min_int), None)

        # 不支持其除 int float 以外的格式
        with self.assertRaises(ValidationError):
            self.assertEqual(tf2.run_validators(data=self.str_min_float), None)
        with self.assertRaises(ValidationError):
            self.assertEqual(tf2.run_validators(data=self.str_min_int), None)

        self.assertEqual(tf2.run_validators(data=self.max_int), None)
        self.assertEqual(tf2.run_validators(data=self.max_float), None)

    def test_max_string_length(self):
        tf1 = TestField(max_digits=6, decimal_places=2)
        with self.assertRaises(ValidationError):
            #  超出约定长度
            tf1.external_to_internal(self.long_str)
        self.assertEqual(tf1.external_to_internal(self.str_max_int), self.max_int)


if __name__ == '__main__':
    unittest.main()



"""
@Author：WangYuXiang
@E-mile：Hill@3io.cc
@CreateTime：2021/2/3 10:07
@DependencyLibrary：无
@MainFunction：无
@FileDoc： 
    test_bool_field.py
    布尔类型字段测试
@ChangeHistory:
    datetime action why
    example:
    2021/2/3 10:07 change 'Fix bug'
        
"""
import asyncio
import unittest
from decimal import Decimal, ROUND_DOWN, getcontext, InvalidOperation

from tortoise.contrib.test import initializer

from sanic_rest_framework.exceptions import ValidationError
from sanic_rest_framework.fields import BooleanField as TestField
from sanic_rest_framework.test.test_fields.test_base_field import TestBaseField

initializer(['sanic_rest_framework.test.models', ],
            # db_url="sqlite://./db.sqlite",
            loop=asyncio.get_event_loop())


class TestDecimalField(TestBaseField):

    def test_external_to_internal(self):
        """
        外转内 str -> dict 是严格的，不符合类型的都应该报错,
        一切都要经过验证
        :return:
        """
        TRUE_VALUES = {
            't', 'T',
            'y', 'Y', 'yes', 'YES',
            'true', 'True', 'TRUE',
            'on', 'On', 'ON',
            '1', 1,
            True
        }
        FALSE_VALUES = {
            'f', 'F',
            'n', 'N', 'no', 'NO',
            'false', 'False', 'FALSE',
            'off', 'Off', 'OFF',
            '0', 0, 0.0,
            False
        }
        NULL_VALUES = {'null', 'Null', 'NULL', '', None}
        tf1 = TestField()
        for i in TRUE_VALUES:
            self.assertEqual(tf1.external_to_internal(i), True)
        for i in FALSE_VALUES:
            self.assertEqual(tf1.external_to_internal(i), False)
        for i in NULL_VALUES:
            self.assertEqual(tf1.external_to_internal(i), None)

    async def test_internal_to_external(self):
        """
        内转外 str -> dict 是宽松的，
        只要类型正确都不报错
        :return:
        """
        TRUE_VALUES = {
            't', 'T',
            'y', 'Y', 'yes', 'YES',
            'true', 'True', 'TRUE',
            'on', 'On', 'ON',
            '1', 1,
            True
        }
        FALSE_VALUES = {
            'f', 'F',
            'n', 'N', 'no', 'NO',
            'false', 'False', 'FALSE',
            'off', 'Off', 'OFF',
            '0', 0, 0.0,
            False
        }
        tf1 = TestField()
        for i in TRUE_VALUES:
            self.assertEqual(await tf1.internal_to_external(i), True)
        for i in FALSE_VALUES:
            self.assertEqual(await tf1.internal_to_external(i), False)
        self.assertEqual(await tf1.internal_to_external(None), False)
        self.assertEqual(await tf1.internal_to_external(''), False)
        self.assertEqual(await tf1.internal_to_external('null'), True)
        self.assertEqual(await tf1.internal_to_external('any'), True)
        self.assertEqual(await tf1.internal_to_external('NULL'), True)
        self.assertEqual(await tf1.internal_to_external('yyyy'), True)
        self.assertEqual(await tf1.internal_to_external('曹凯'), True)


if __name__ == '__main__':
    unittest.main()



"""
@Author：WangYuXiang
@E-mile：Hill@3io.cc
@CreateTime：2021/3/4 11:20
@DependencyLibrary：无
@MainFunction：无
@FileDoc： 
    test_choice_field.py
    文件说明
@ChangeHistory:
    datetime action why
    example:
    2021/3/4 11:20 change 'Fix bug'
        
"""

import unittest

from sanic_rest_framework.exceptions import ValidationError
from sanic_rest_framework.fields import ChoiceField as TestField
from sanic_rest_framework.test.test_fields.test_base_field import TestBaseField


class TestDateTimeField(TestBaseField):
    def setUp(self) -> None:
        self.choice = (
            ('刘文静', '开发人员'),
            ('光明', '播音员'),
            ('李焕英', '纺织厂女工')
        )

    def test_external_to_internal(self):
        """
        外转内 str -> dict 是严格的，不符合类型的都应该报错,
        一切都要经过验证
        :return:
        """

        # 正常测试
        tf1 = TestField(choices=self.choice)
        self.assertEqual(tf1.external_to_internal('刘文静'), '开发人员')
        self.assertEqual(tf1.external_to_internal('光明'), '播音员')
        self.assertEqual(tf1.external_to_internal('李焕英'), '纺织厂女工')

        with self.assertRaises(ValidationError):
            tf1.external_to_internal(None)
        with self.assertRaises(ValidationError):
            tf1.external_to_internal('曹凯')

    async def test_internal_to_external(self):
        """
        内转外 str -> dict 是宽松的，
        只要类型正确都不报错
        :return:
        """
        # 正常测试
        tf1 = TestField(choices=self.choice)
        self.assertEqual(await tf1.internal_to_external('刘文静'), '开发人员')
        self.assertEqual(await tf1.internal_to_external('光明'), '播音员')
        self.assertEqual(await tf1.internal_to_external('李焕英'), '纺织厂女工')

        with self.assertRaises(ValidationError):
            await tf1.internal_to_external(None)
        with self.assertRaises(ValidationError):
            await tf1.internal_to_external('曹凯')


if __name__ == '__main__':
    unittest.main()
